// Copyright (c) 2009-2011 Intel Corporation
// All rights reserved.
// 
// WARRANTY DISCLAIMER
// 
// THESE MATERIALS ARE PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL INTEL OR ITS
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
// OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THESE
// MATERIALS, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
// Intel Corporation is the author of the Materials, and requests that all
// problem reports or change requests be submitted to it directly

#include "stdafx.h"
#include "CL\cl.h"
#include "utils.h"
#include "Bitonic_APU.h"

cl_mem		g_inputBuffer = NULL;
cl_context	g_context = NULL;
cl_command_queue g_cmd_queue = NULL;
cl_program	g_program = NULL;
cl_kernel	g_kernel = NULL;

bool g_bRunOnPG = false;

#if (BITONIC_USE_APU == 1)
const char kernel_file_name[256] = "Bitonic_APU.cl";
const char kernel_function_name[256] = "BitonicSort_WG";
#else
const char kernel_file_name[256] = "BitonicSort.cl";
const char kernel_function_name[256] = "BitonicSort";
#endif

//for perf. counters
#include <Windows.h>

LARGE_INTEGER g_PerfFrequency;
LARGE_INTEGER g_PerformanceCountNDRangeStart;
LARGE_INTEGER g_PerformanceCountNDRangeStop;
LARGE_INTEGER g_PerformanceCountReferenceStart;
LARGE_INTEGER g_PerformanceCountReferenceStop;

void Cleanup_OpenCL()
{
    if( g_inputBuffer ) {clReleaseMemObject( g_inputBuffer ); g_inputBuffer = NULL;}
    if( g_kernel ) {clReleaseKernel( g_kernel ); g_kernel = NULL;}
    if( g_program ) {clReleaseProgram( g_program ); g_program = NULL;}
    if( g_cmd_queue ) {clReleaseCommandQueue( g_cmd_queue ); g_cmd_queue = NULL;}
    if( g_context ) {clReleaseContext( g_context ); g_context = NULL;}
}

int Setup_OpenCL( const char *program_source, cl_uint* alignment)
{
    cl_device_id devices[16];
    size_t cb;
    cl_uint size_ret = 0;
    cl_int err;

	if(g_bRunOnPG)
	{
		printf("Trying to run on a Processor Graphics \n");
	}
	else
	{
		printf("Trying to run on a CPU \n");
	}

	cl_platform_id intel_platform_id = GetIntelOCLPlatform();
    if( intel_platform_id == NULL )
    {
        printf("ERROR: Failed to find Intel OpenCL platform.\n");
        return -1;
    }

    cl_context_properties context_properties[3] = {CL_CONTEXT_PLATFORM, (cl_context_properties)intel_platform_id, NULL };

    // create the OpenCL context on a CPU/PG 
	if(g_bRunOnPG)
	{
		g_context = clCreateContextFromType(context_properties, CL_DEVICE_TYPE_GPU, NULL, NULL, NULL);
	}
	else
	{
		g_context = clCreateContextFromType(context_properties, CL_DEVICE_TYPE_CPU, NULL, NULL, NULL);
	}
    if (g_context == (cl_context)0)
        return -1;


    // get the list of CPU devices associated with context
    err = clGetContextInfo(g_context, CL_CONTEXT_DEVICES, 0, NULL, &cb);
    clGetContextInfo(g_context, CL_CONTEXT_DEVICES, cb, devices, NULL);

    if( alignment )
    {
        err = clGetDeviceInfo (devices[0],
            CL_DEVICE_MEM_BASE_ADDR_ALIGN ,
            sizeof(cl_uint),
            alignment,
            NULL);

        *alignment/=8; //in bytes
        printf("OpenCL data alignment is %d bytes.\n", *alignment);
    }

    g_cmd_queue = clCreateCommandQueue(g_context, devices[0], 0, NULL);
    if (g_cmd_queue == (cl_command_queue)0)
    {
        Cleanup_OpenCL();
        return -1;
    }

    char *sources = ReadSources(program_source);	//read program .cl source file
	if( NULL == sources )
	{
        printf("ERROR: Failed to read sources into memory...\n");
        Cleanup_OpenCL();
        return -1;
    }

    g_program = clCreateProgramWithSource(g_context, 1, (const char**)&sources, NULL, NULL);
    if (g_program == (cl_program)0)
    {
        printf("ERROR: Failed to create Program with source...\n");
        Cleanup_OpenCL();
        free(sources);
        return -1;
    }

	err = clBuildProgram(g_program, 
		                 0, 
						 NULL, 
#if (OPENCL_KERNEL_DEBUG == 1)
						 "-g -s \"D:/Work/Program/Window/OpenCL/Bitonic_APU/Bitonic_APU/Bitonic_APU/Bitonic_APU.cl\"", 
#else
						 NULL, 
#endif
						 NULL,
						 NULL);
    if (err != CL_SUCCESS)
    {
        printf("ERROR: Failed to build program...\n");
        BuildFailLog(g_program, devices[0]);
        Cleanup_OpenCL();
        free(sources);
        return -1;
    }
#if (BITONIC_USE_APU == 1)
	g_kernel = clCreateKernel(g_program, "BitonicSort_WG_2", NULL);
	//g_kernel = clCreateKernel(g_program, "TEST_BARRIER_0", NULL);   
	//g_kernel = clCreateKernel(g_program, "BitonicSort_WG", NULL);    
#else
	g_kernel = clCreateKernel(g_program, "BitonicSort", NULL);
#endif

	if (g_kernel == (cl_kernel)0)
    {
        printf("ERROR: Failed to create kernel...\n");
        Cleanup_OpenCL();
        free(sources);
        return -1;
    }
    free(sources);

    return 0; // success...
}

void generateInput(cl_int* inputArray, size_t arraySize)
{
    const size_t minElement = 0;
    const size_t maxElement = arraySize + 1;

    srand(12345);

    // random initialization of input
    for (size_t i = 0; i < arraySize; ++i)
    {
        inputArray[i] = (cl_int)((float) (maxElement - minElement) * (rand() / (float) RAND_MAX));
    }

#ifdef BITONIC_DEBUG_FILE_OUT
	FILE *pFile0;
	char fileName0[256];// = "e:\\temp\\pc_sim_180x148_yuv422_rot3.raw";

	sprintf(fileName0, "d:\\temp\\inputArray.bin");

	pFile0 = fopen((const char*)fileName0, "w+b");
	fwrite((const void*)inputArray, sizeof(char), arraySize*4, pFile0);
	fclose(pFile0);
#endif
}

void ExecuteSortReference(cl_int* inputArray, cl_int arraySize, cl_bool sortAscending)
{
    cl_int numStages = 0;
    cl_uint temp;

    cl_int stage;
    cl_int passOfStage;

	QueryPerformanceCounter(&g_PerformanceCountReferenceStart);

    for (temp = arraySize; temp > 1; temp >>= 1)
    {
        ++numStages;
    }

    for (stage = 0; stage < numStages; ++stage)
    {
        int dirMask = 1 << stage;

        for (passOfStage = 0; passOfStage < stage + 1; ++passOfStage)
        {
            const cl_uint shift = stage - passOfStage;
            const cl_uint distance = 1 << shift;
            const cl_uint lmask = distance - 1;

            for(int g_id=0; g_id < arraySize >> 1; g_id++)
            {
                cl_uint leftId = (( g_id >> shift ) << (shift + 1)) + (g_id & lmask);
                cl_uint rightId = leftId + distance;

                cl_uint left  = inputArray[leftId];
                cl_uint right = inputArray[rightId];

                cl_uint greater;
                cl_uint lesser;

                if(left > right)
                {
                    greater = left;
                    lesser  = right;
                }
                else
                {
                    greater = right;
                    lesser  = left;
                }

                cl_bool dir = sortAscending;
                if( ( g_id & dirMask) == dirMask )
                    dir = !dir;

                if(dir)
                {
                    inputArray[leftId]  = lesser;
                    inputArray[rightId] = greater;
                }
                else
                {
                    inputArray[leftId]  = greater;
                    inputArray[rightId] = lesser;
                }
            }

#ifdef BITONIC_DEBUG_FILE_OUT
			FILE *pFile0;
			char fileName0[256];// = "e:\\temp\\pc_sim_180x148_yuv422_rot3.raw";

			sprintf(fileName0, "d:\\temp\\Stage%d_passOfStage%d_ref.bin", 
			        stage, 
			        passOfStage);

			pFile0 = fopen((const char*)fileName0, "w+b");
			fwrite((const void*)inputArray, sizeof(char), arraySize*4, pFile0);
			fclose(pFile0);
#endif
        }
    }
	QueryPerformanceCounter(&g_PerformanceCountReferenceStop);
}

bool ExecuteSortKernel(cl_int* inputArray, cl_int arraySize, cl_uint sortAscending)
{
    cl_int err = CL_SUCCESS;
    cl_int numStages = 0;
    cl_uint temp;

    cl_int stage;
    cl_int passOfStage;

    //create OpenCL buffer using input array memory
    g_inputBuffer = clCreateBuffer(g_context, CL_MEM_USE_HOST_PTR, sizeof(cl_int) * arraySize, inputArray, NULL);

    if (g_inputBuffer == (cl_mem)0)
    {
        printf("ERROR: Failed to create input data Buffer\n");
        return false;
    }

    for (temp = arraySize; temp > 2; temp >>= 1)
    {
        numStages++;
    }

    err  = clSetKernelArg(g_kernel, 0, sizeof(cl_mem), (void *) &g_inputBuffer);
    err |= clSetKernelArg(g_kernel, 3, sizeof(cl_uint), (void *) &sortAscending);
    if (err != CL_SUCCESS)
    {
        printf("ERROR: Failed to set input kernel arguments\n");
        return false;
    }

    QueryPerformanceCounter(&g_PerformanceCountNDRangeStart);
    for (stage = 0; stage < numStages; stage++)
    {
        if ( CL_SUCCESS != clSetKernelArg(g_kernel, 1, sizeof(cl_uint), (void *) &stage) )
            return false;

        for (passOfStage = stage; passOfStage >= 0; passOfStage--)
        {
            if ( CL_SUCCESS != clSetKernelArg(g_kernel, 2, sizeof(cl_uint), (void *) &passOfStage) )
                return false;

#ifdef BITONIC_REFERENCE_FILE_OUT
			stage =         7;//DESCENDING --> 6;// 5;// 1;// 0;// 7;// //ASCENDING --> 5;//8;//0;//7;//8;//
			passOfStage =   3;//DESCENDING --> 6;// 4;// 0;// 0;// 3;// //ASCENDING --> 1;//8;//0;//0;//7;//
			if ( CL_SUCCESS != clSetKernelArg(g_kernel, 1, sizeof(cl_uint), (void *) &stage) )
				return false;

			if ( CL_SUCCESS != clSetKernelArg(g_kernel, 2, sizeof(cl_uint), (void *) &passOfStage) )
                return false;
#endif
            // set work-item dimensions
            size_t gsz = arraySize / (2*4);
            size_t global_work_size[1] = { passOfStage ? gsz : gsz << 1 };	//number of quad items in input array
            size_t local_work_size[1]= { BITONIC_LOCAL_SIZE };					//valid WG sizes are 1:1024

            // execute kernel
            if (CL_SUCCESS != clEnqueueNDRangeKernel(g_cmd_queue, g_kernel, 1, NULL, global_work_size, local_work_size, 0, NULL, NULL))
            {
                printf("ERROR: Failed to execute sorting kernel\n");
                return false;
            }
#if defined(BITONIC_DEBUG_FILE_OUT) || defined(BITONIC_REFERENCE_FILE_OUT)

			err = clFinish(g_cmd_queue);

			void* tmp_ptr = NULL;
			tmp_ptr = clEnqueueMapBuffer(g_cmd_queue, g_inputBuffer, true, CL_MAP_READ, 0, arraySize , 0, NULL, NULL, NULL);
			if(tmp_ptr!=inputArray)
			{
				printf("ERROR: clEnqueueMapBuffer failed to return original pointer\n");
				return false;
			}

			err = clFinish(g_cmd_queue);

			FILE *pFile0;
			char fileName0[256];// = "e:\\temp\\pc_sim_180x148_yuv422_rot3.raw";

			sprintf(fileName0, "d:\\temp\\Stage%d_passOfStage%d_cl.bin", 
			        stage, 
			        passOfStage);

			pFile0 = fopen((const char*)fileName0, "w+b");
			fwrite((const void*)inputArray, sizeof(char), arraySize*4, pFile0);
			fclose(pFile0);
#endif
        }
    }

    err = clFinish(g_cmd_queue);
    QueryPerformanceCounter(&g_PerformanceCountNDRangeStop);

    void* tmp_ptr = NULL;
    tmp_ptr = clEnqueueMapBuffer(g_cmd_queue, g_inputBuffer, true, CL_MAP_READ, 0, sizeof(cl_int) * arraySize , 0, NULL, NULL, NULL);
	if(tmp_ptr!=inputArray)
	{
		printf("ERROR: clEnqueueMapBuffer failed to return original pointer\n");
		return false;
	}


    err = clFinish(g_cmd_queue);

    clEnqueueUnmapMemObject(g_cmd_queue, g_inputBuffer, tmp_ptr, 0, NULL, NULL);

    clReleaseMemObject(g_inputBuffer); g_inputBuffer = NULL;

    return true;
}

void Usage()
{
    printf("Usage: BitonicSort.exe [--h] [-s <arraySize>] [-d]\n");
    printf("  where, --h prints this message\n");
    printf("    <arraySize> is input/output array size\n");
    printf("    -d performs descending sort (default is ascending)\n");
    printf("    -g run on Processor Graphics\n");
    exit(-1);
}

int _tmain(int argc, _TCHAR* argv[])
{
    cl_uint dev_alignment = BITONIC_LOCAL_SIZE;
    cl_bool sortAscending = BITONIC_ASENDING_SORT;
    cl_int arraySize = BITONIC_ARRAY_SIZE;


    int argn = 1;
    while (argn < argc)
    {
        if (_tcscmp(argv[argn], _T("--h")) == 0)
        {
            Usage();
        }
        else if (_tcscmp(argv[argn], _T("-s")) == 0)
        {
            if(++argn==argc)
                Usage();
            arraySize = _ttoi(argv[argn]);
            argn++;
        }
        else if (_tcscmp(argv[argn], _T("-d")) == 0)
        {
            sortAscending = false;
            argn++;
        }
        else if (_tcscmp(argv[argn], _T("-g")) == 0)
        {
            g_bRunOnPG = true;
            argn++;
        }
        else
        {
            argn++;
        }
    }
    if( argc < 2 )
    {
        printf("No command line arguments specified, using default values.\n");
    }

    //validate user input parameters, if any
    {
        cl_uint sz_log_b = 0;

        if( arraySize < 1024 )
        {
            printf("Input size should be no less than 1024!\n");
            return -1;
        }

        for (int temp = arraySize; temp>1; temp >>= 1)
            sz_log_b++;

        if( arraySize & ((1 << sz_log_b)-1)  )
        {
            printf("Input size should be (2^N)*4!\n");
            return -1;
        }
    }

    printf("Initializing OpenCL runtime...\n");

    //initialize Open CL objects (context, queue, etc.)
    if( 0 != Setup_OpenCL(kernel_file_name, &dev_alignment) )
        return -1;

    printf("Sort order is %s\n", sortAscending ? "ascending" : "descending" );
    printf("Input size is %d items\n", arraySize);
    printf("Work grop size %d \n", BITONIC_LOCAL_SIZE);    
	cl_int* inputArray = (cl_int*)_aligned_malloc(sizeof(cl_int) * arraySize, dev_alignment);
	//cl_int* inputArray = (cl_int*)_aligned_malloc(sizeof(cl_int) * arraySize*4, dev_alignment);
    cl_int* refArray = (cl_int*)_aligned_malloc(sizeof(cl_int) * arraySize, dev_alignment);

	//memset((void*)inputArray, 0x0, sizeof(cl_int) * arraySize*4);

    //random input
    generateInput(inputArray, arraySize);
    memcpy(refArray, inputArray, sizeof(cl_int) * arraySize);

    //sort input array in a given direction
    printf("Executing reference...\n");
    ExecuteSortReference(refArray, arraySize, sortAscending);

    printf("Executing OpenCL kernel...\n");
    ExecuteSortKernel(inputArray, arraySize, sortAscending);


    bool result = true;
    printf("Performing verification...\n");
    for (cl_int i = 0; i < arraySize; i++)
    {
        if (inputArray[i] != refArray[i])
        {
            result = false;
            //break;
        }
    }

    if(!result)
    {
        printf("ERROR: Verification failed.\n");
    }
    else
    {
        printf("******* Verification Succeeded. ******* \n");
    }

    //retrieve perf. counter frequency
    QueryPerformanceFrequency(&g_PerfFrequency);

#if (BITONIC_USE_APU == 1)
	printf("modified ");
#else
	printf("original ");
#endif

    printf("OpenCL code execution time %f ms.\n", 
        1000.0f*(float)(g_PerformanceCountNDRangeStop.QuadPart - g_PerformanceCountNDRangeStart.QuadPart)/(float)g_PerfFrequency.QuadPart);

    printf("Reference code execution time %f ms.\n", 
        1000.0f*(float)(g_PerformanceCountReferenceStop.QuadPart - g_PerformanceCountReferenceStart.QuadPart)/(float)g_PerfFrequency.QuadPart);

    printf("Releasing resources...\n");
    _aligned_free( refArray );
    _aligned_free( inputArray );

    Cleanup_OpenCL();

    return 0;
}

